@@ -179,13 +179,11 @@ def normalize_strip(content: str) -> str:
 
 FULL_NAME_DOC_REMAPS = {
     'builtins.input': '''
-input(prompt: Any=...) -> str
+input(prompt: Any=...) -> Optional[str]
 
-Prompt the user to input a string (which might be empty).
-If the prompt is closed or canceled, `None` is returned.
-
-Note that calling this function will pause the turtle simulation
-while the user decides what to enter.
+Prompt the user to input a string.
+If the prompt is closed or canceled, `None` is returned,
+otherwise the user's input is returned directly (which may be the empty string).
 '''.strip(),
 }
 PROP_DOC_REMAPS = {}
@@ -347,7 +345,7 @@ class DndManager:
         widget.bind('<ButtonPress-1>', self.on_start)
         widget.bind('<B1-Motion>', self.on_drag)
         widget.bind('<ButtonRelease-1>', self.on_drop)
-    
+
     def on_start(self, e):
         for target in self.targets:
             target.on_start(e)
@@ -395,7 +393,7 @@ class BlocksList(tk.Frame):
                 # restore saved focus
                 if focused[0] is not None:
                     focused[0].focus()
-    
+
                 text_target.configure(highlightbackground = orig_bcolor)
             def on_drop(e):
                 x, y = text_target.winfo_pointerxy()
@@ -447,6 +445,25 @@ class Imports:
              - `random.randrange(a,b)` gives you a random integer in `range(a,b)`.
             Just like the `range` function, you can pass a single argument to default the first to 0.
         '''],
+        ['fractions', 'frac', '''
+            The `fractions` module (`frac` for short) allows you to do math directly on rational numbers (fractions).
+
+            If you use built-in types like `float` to do numeric calculations, you might encounter some amount of rounding error.
+            This is because `float` has a maximum number of "significant digits" it can represent, meaning it cannot store all numbers exactly.
+            For instance, the number `1/3` written in binary has an infinite repeating patterns of 1's and 0's, so `float` cannot store it exactly.
+            On the other hand, `frac.Fraction` stores the number as a pair of integers (numerator and denominator), so fractions can be stored exactly.
+
+            To create a fraction, you can use `frac.Fraction(my_number)` to perform the conversion.
+            Fractions support the typical math operators, so you can write (frac.Fraction(773) / frac.Fraction(3))**2 and it will give you
+            back a fraction containing exactly 597529/9 (the same calculation using `float` has an error of `1.4552e-11`).
+            Although rounding error from individual operations on `float` are small, they add up when performing many calculations.
+            `frac.Fraction` can be used to avoid any error, so long as you only use operations that work with rational numbers
+            (e.g., you cannot take the cos of a fraction, but you can add, subtract, multiply, divide, raise to integer powers, etc.).
+
+            Printing out a fraction will show the value in fraction form.
+            If you want to convert the fraction into a `float` (e.g., to display it as a decimal),
+            you can use `float(my_frac)` to perform the conversion.
+        '''],
         ['numpy', 'np', '''
             The `numpy` package is a highly specialized tool for doing a lot of number crunching very very quickly.
             By convention, `numpy` is typically imported as `np` for brevity.
@@ -481,7 +498,7 @@ class Imports:
                 'info': clean_docstring(normalize_strip(item[2])),
                 'code': f'import {item[0]}' if item[0] == item[1] else f'import {item[0]} as {item[1]}',
             }
-        
+
         self.images = {}
 
         self.on_update = on_update
@@ -512,7 +529,7 @@ class Imports:
             import_str = '\n'.join(lines)
             GlobalEditor.prefix = f'{GlobalEditor.BASE_PREFIX}{import_str}\n\n'
             GlobalEditor.prefix_lines = GlobalEditor.BASE_PREFIX_LINES + len(lines) + 1
-        
+
         if self.on_update is not None:
             self.on_update()
 
@@ -524,7 +541,6 @@ class ProjectEditor(tk.Frame):
         # it is not stored in project files or exports because conflicting ids would break messaging.
         self.project_id = common.generate_proj_id()
 
-        self.turtle_index = 0
         self.editors: List[CodeEditor] = []
         self.show_blocks_tkvar = tk.BooleanVar()
 
@@ -562,9 +578,9 @@ class ProjectEditor(tk.Frame):
             self.ctx_menu.add_command(label = label, command = command)
             idx = len(self.ctx_menu_entries)
             self.ctx_menu_entries[id] = idx
-        
-        add_command('new-turtle', label = 'New Turtle', command = lambda: self.newturtle())
-        add_command('dupe', label = 'Clone Turtle', command = lambda: self.dupe_turtle(self.ctx_tab_idx))
+
+        add_command('new-turtle', label = 'New Sprite', command = lambda: self.newturtle())
+        add_command('dupe', label = 'Clone Sprite', command = lambda: self.dupe_turtle(self.ctx_tab_idx))
         add_command('rename', label = 'Rename', command = lambda: self.rename_turtle(self.ctx_tab_idx))
         add_command('delete', label = 'Delete', command = lambda: self.delete_tab(self.ctx_tab_idx))
 
@@ -616,7 +632,7 @@ class ProjectEditor(tk.Frame):
         if not isinstance(editor, TurtleEditor):
             return # only turtle editors can be duped
 
-        return self.newturtle(base_name = editor.name, value = editor.text.get('1.0', 'end-1c'))
+        return self.newturtle(base_name = f'{editor.name}_copy', value = editor.text.get('1.0', 'end-1c'))
 
     def rename_turtle(self, idx) -> None:
         editor = self.editors[idx]
@@ -625,8 +641,8 @@ class ProjectEditor(tk.Frame):
 
         name = None
         while True:
-            title = 'Rename Turtle'
-            msg = f'Enter the new name for turtle "{editor.name}" (must not already be taken).\nNote that any references to this turtle in your code must be manually updated to the new name.'
+            title = 'Rename Sprite'
+            msg = f'Enter the new name for "{editor.name}" (must not already be taken).\nNote that any references to this sprite in your code must be manually updated to the new name.'
             name = simpledialog.askstring(title, msg)
             if name is None:
                 return
@@ -637,23 +653,23 @@ class ProjectEditor(tk.Frame):
                 messagebox.showerror(title = 'Invalid name', message = f'A tab named "{name}" already exists')
                 continue
             break
-        
+
         editor.name = name
         self.notebook.tab(idx, text = name)
 
-    def newturtle(self, *, base_name = 'turtle', value = None) -> Any:
-        while True:
-            self.turtle_index += 1
-            name = f'{base_name}{self.turtle_index}'
-            if self.is_unique_name(name):
-                break
+    def newturtle(self, *, base_name = 'sprite', value = None) -> Any:
+        name_counter = 0
+        name = base_name
+        while not self.is_unique_name(name):
+            name_counter += 1
+            name = f'{base_name}{name_counter}'
 
         assert self.is_unique_name(name) and is_valid_ident(name) # sanity check
         editor = TurtleEditor(self.notebook, name = name, value = value or ProjectEditor.DEFAULT_PROJECT['editors'][2]['value'])
         self.notebook.add(editor, text = name)
         self.editors.append(editor)
         return editor
-    
+
     def get_full_script(self, *, is_export: bool = False) -> str:
         scripts = []
         for editor in self.editors:
@@ -663,8 +679,10 @@ class ProjectEditor(tk.Frame):
         return ''.join(scripts)
 
     DEFAULT_GLOBAL_BLOCKS = [
-        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef my_onstart(): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart()\ndef my_onstart(): # functions need different names\n    pass # replace with your code' },
         { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef my_onkey(): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/msgrecv-snap.png', 'scale': 1, 'replace': '@nb.on_message(\'message\')\ndef my_on_message(msg): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/sendmsg-snap.png', 'scale': 1, 'replace': 'nb.send_message(\'message\', \'myself\', msg = \'Hello World!\')' },
         { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass # replace with your code' },
 
         { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass # replace with your code' },
@@ -674,9 +692,11 @@ class ProjectEditor(tk.Frame):
         { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in my_list:\n    pass # replace with your code' },
     ]
     DEFAULT_STAGE_BLOCKS = [
-        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef my_onstart(self): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart()\ndef my_onstart(self): # functions need different names\n    pass # replace with your code' },
         { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef my_onkey(self): # functions need different names\n    pass # replace with your code' },
-        { 'url': f'{IMG_ROOT}/blocks/onclick.png', 'scale': 1, 'replace': '@onclick\ndef my_onclick(self, x, y): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onclick.png', 'scale': 1, 'replace': '@onclick()\ndef my_onclick(self, x, y): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/msgrecv-snap.png', 'scale': 1, 'replace': '@nb.on_message(\'message\')\ndef my_on_message(self, msg): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/sendmsg-snap.png', 'scale': 1, 'replace': 'nb.send_message(\'message\', \'myself\', msg = \'Hello World!\')' },
         { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass # replace with your code' },
 
         { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass # replace with your code' },
@@ -686,14 +706,16 @@ class ProjectEditor(tk.Frame):
         { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in my_list:\n    pass # replace with your code' },
     ]
     DEFAULT_TURTLE_BLOCKS = [
-        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart\ndef my_onstart(self): # functions need different names\n    pass # replace with your code' },
-        { 'url': f'{IMG_ROOT}/blocks/onstartclone.png', 'scale': 1, 'replace': '@onstartclone\ndef my_onstartclone(self): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onstart.png', 'scale': 1, 'replace': '@onstart()\ndef my_onstart(self): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onstartclone.png', 'scale': 1, 'replace': '@onstart(when = \'clone\')\ndef my_onstartclone(self): # functions need different names\n    pass # replace with your code' },
         { 'url': f'{IMG_ROOT}/blocks/keypress.png', 'scale': 1, 'replace': '@onkey(\'space\')\ndef my_onkey(self): # functions need different names\n    pass # replace with your code' },
-        { 'url': f'{IMG_ROOT}/blocks/onclick.png', 'scale': 1, 'replace': '@onclick\ndef my_onclick(self, x, y): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/onclick.png', 'scale': 1, 'replace': '@onclick()\ndef my_onclick(self, x, y): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/msgrecv-snap.png', 'scale': 1, 'replace': '@nb.on_message(\'message\')\ndef my_on_message(self, msg): # functions need different names\n    pass # replace with your code' },
+        { 'url': f'{IMG_ROOT}/blocks/sendmsg-snap.png', 'scale': 1, 'replace': 'nb.send_message(\'message\', \'myself\', msg = \'Hello World!\')' },
         { 'url': f'{IMG_ROOT}/blocks/loop-forever.png', 'scale': 1, 'replace': 'while True:\n    pass # replace with your code' },
 
         { 'url': f'{IMG_ROOT}/blocks/warp.png', 'scale': 1, 'replace': 'with Warp():\n    pass # replace with your code' },
-        
+
         { 'url': f'{IMG_ROOT}/blocks/msgrecv.png', 'scale': 1, 'replace': '@nb.on_message(\'message\')\ndef my_on_message(self, msg): # functions need different names\n    pass # replace with your code' },
         { 'url': f'{IMG_ROOT}/blocks/sendmsg.png', 'scale': 1, 'replace': 'nb.send_message(\'message\', \'myself\', msg = \'Hello World!\')' },
         { 'url': f'{IMG_ROOT}/blocks/loop-foreach.png', 'scale': 1, 'replace': 'for item in my_list:\n    pass # replace with your code' },
@@ -703,7 +725,6 @@ class ProjectEditor(tk.Frame):
         'stage_blocks': DEFAULT_STAGE_BLOCKS,
         'turtle_blocks': DEFAULT_TURTLE_BLOCKS,
         'show_blocks': True,
-        'turtle_index': 0,
         'editors': [
             {
                 'type': 'global',
@@ -717,27 +738,25 @@ my_global_var = 'hello world' # create a global variable
                 'name': 'stage',
                 'value': '''
 @onstart
-def start(self):
+def my_onstart(self): # functions need different names
     self.my_list = [1, 2, 3] # create a stage variable
 '''.lstrip(),
             },
             {
                 'type': 'turtle',
-                'name': 'turtle',
+                'name': 'sprite',
                 'value': '''
 @onstart
-def start(self):
-    self.my_distance = 100 # create a turtle variable
+def my_onstart(self): # functions need different names
+    self.my_distance = 2 # create a sprite variable
 
-    for i in range(4): # repeat code 4 times
-        time.sleep(0.5) # wait half a second
+    for i in range(360): # repeat code 360 times
         self.forward(self.my_distance)
-        time.sleep(0.5)
-        self.turn_right()
+        self.turn_right(1)
 '''.lstrip(),
             },
         ],
-        'imports': ['time'],
+        'imports': [],
     }
 
     def get_save_dict(self) -> dict:
@@ -746,7 +765,6 @@ def start(self):
         res['stage_blocks'] = [x.copy() for x in  StageEditor.blocks]
         res['turtle_blocks'] = [x.copy() for x in TurtleEditor.blocks]
         res['show_blocks'] = self.show_blocks
-        res['turtle_index'] = self.turtle_index
         res['imports'] = []
         for pkg, item in self.imports.packages.items():
             if item['tkvar'].get():
@@ -777,7 +795,6 @@ def start(self):
 
         self.show_blocks = proj['show_blocks']
 
-        self.turtle_index = proj['turtle_index']
         for info in proj['editors']:
             ty = info['type']
             name = info['name']
@@ -889,16 +906,22 @@ class ChangedText(tk.Text):
         return result # return what the actual widget returned
 
 class ScrolledText(tk.Frame):
-    def __init__(self, parent, *, readonly = False, linenumbers = False, blocks = [], **kwargs):
+    def __init__(self, parent, *, wrap = True, readonly = False, linenumbers = False, blocks = [], **kwargs):
         super().__init__(parent)
         undo_args = { 'undo': True, 'maxundo': -1, 'autoseparators': True }
 
         self.font = tkfont.nametofont('TkFixedFont')
 
         self.scrollbar = ttk.Scrollbar(self)
-        self.text = ChangedText(self, font = self.font, yscrollcommand = self.scrollbar.set, **({} if readonly else undo_args), **kwargs)
+        self.text = ChangedText(self, font = self.font, yscrollcommand = self.scrollbar.set, wrap = tk.WORD if wrap else tk.NONE, **({} if readonly else undo_args), **kwargs)
         self.scrollbar.config(command = self.text.yview)
 
+        self.hscrollbar = None
+        if not wrap:
+            self.hscrollbar = ttk.Scrollbar(self, orient = 'horizontal')
+            self.text.config(xscrollcommand = self.hscrollbar.set)
+            self.hscrollbar.config(command = self.text.xview)
+
         self.custom_on_change = []
 
         def on_select_all(e):
@@ -945,7 +968,7 @@ class ScrolledText(tk.Frame):
             def on_paste(e):
                 if self.text.tag_ranges(tk.SEL):
                     self.text.delete(tk.SEL_FIRST, tk.SEL_LAST)
-                
+
                 # some versions of tcl/tk on mac are broken and crash here, so impl manually
                 self.text.insert(tk.INSERT, self.clipboard_get())
                 return 'break'
@@ -965,7 +988,7 @@ class ScrolledText(tk.Frame):
         self.update_pack()
 
     def update_pack(self):
-        for item in [self.scrollbar, self.blocks, self.linenumbers, self.text]:
+        for item in [self.scrollbar, self.hscrollbar, self.blocks, self.linenumbers, self.text]:
             if item is not None: item.pack_forget()
 
         self.scrollbar.pack(side = tk.RIGHT, fill = tk.Y)
@@ -973,6 +996,8 @@ class ScrolledText(tk.Frame):
             self.blocks.pack(side = tk.LEFT, fill = tk.Y)
         if self.linenumbers is not None:
             self.linenumbers.pack(side = tk.LEFT, fill = tk.Y)
+        if self.hscrollbar is not None:
+            self.hscrollbar.pack(side = tk.BOTTOM, fill = tk.X)
         self.text.pack(side = tk.RIGHT, fill = tk.BOTH, expand = True)
 
         self.update() # needed on mac
@@ -988,8 +1013,8 @@ class ScrolledText(tk.Frame):
         self.text.insert('1.0', txt)
 
 class CodeEditor(ScrolledText):
-    def __init__(self, parent, *, column_offset = 0, **kwargs):
-        super().__init__(parent, linenumbers = True, **kwargs)
+    def __init__(self, parent, *, wrap = False, column_offset = 0, **kwargs):
+        super().__init__(parent, linenumbers = True, wrap = wrap, **kwargs)
         self.__line_count = None
         self.column_offset = column_offset
         self.help_popup = None
@@ -1040,15 +1065,15 @@ class CodeEditor(ScrolledText):
             ]
             cdg.prog = re.compile('|'.join(patterns))
 
-            cdg.tagdefs['COMMENT']    = {'foreground': '#a3a3a3', 'background': COLOR_INFO['text-background']}
-            cdg.tagdefs['MYNUMBER']   = {'foreground': '#c26910', 'background': COLOR_INFO['text-background']}
-            cdg.tagdefs['MYSELF']     = {'foreground': '#d943aa', 'background': COLOR_INFO['text-background']}
-            cdg.tagdefs['MYPROP']     = {'foreground': '#d943aa', 'background': COLOR_INFO['text-background']}
-            cdg.tagdefs['BUILTIN']    = {'foreground': '#6414b5', 'background': COLOR_INFO['text-background']}
-            cdg.tagdefs['DEFINITION'] = {'foreground': '#6414b5', 'background': COLOR_INFO['text-background']}
-            cdg.tagdefs['MYDECO']     = {'foreground': '#6414b5', 'background': COLOR_INFO['text-background']}
-            cdg.tagdefs['KEYWORD']    = {'foreground': '#0d15b8', 'background': COLOR_INFO['text-background']}
-            cdg.tagdefs['STRING']     = {'foreground': '#961a1a', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['COMMENT']    = {'foreground': '#a3a3a3' if IS_DARK else '#a3a3a3', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['MYNUMBER']   = {'foreground': '#e8821c' if IS_DARK else '#c26910', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['MYSELF']     = {'foreground': '#fc72d0' if IS_DARK else '#d943aa', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['MYPROP']     = {'foreground': '#fc72d0' if IS_DARK else '#d943aa', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['BUILTIN']    = {'foreground': '#b576f5' if IS_DARK else '#6414b5', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['DEFINITION'] = {'foreground': '#b576f5' if IS_DARK else '#6414b5', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['MYDECO']     = {'foreground': '#b576f5' if IS_DARK else '#6414b5', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['KEYWORD']    = {'foreground': '#5b96f5' if IS_DARK else '#0d15b8', 'background': COLOR_INFO['text-background']}
+            cdg.tagdefs['STRING']     = {'foreground': '#f24141' if IS_DARK else '#961a1a', 'background': COLOR_INFO['text-background']}
 
             percolator.Percolator(self.text).insertfilter(cdg)
 
@@ -1254,7 +1279,7 @@ class CodeEditor(ScrolledText):
             self.text.insert(tk.INSERT, '    ')
 
         return 'break' # we always override default (we don't want tabs ever)
-    
+
     def do_autocomment(self):
         self._do_batch_edit(smart_comment_uncomment)
         return 'break'
@@ -1266,8 +1291,7 @@ from netsblox.turtle import *
 from netsblox.concurrency import *
 nb = netsblox.Client(proj_name = """$proj_name""", proj_id = $proj_id)
 'A connection to NetsBlox, which allows you to use services and RPCs from python.'
-netsblox.turtle._set_window_size(1280, 720)
-netsblox.turtle._turtle.title(f'NetsBlox-Python - {nb.get_public_id()}')
+getattr(netsblox.turtle._get_proj_handle(), '_Project__tk').title(f'PyBlox - {nb.get_public_id()}')
 setup_stdio()
 setup_yielding()
 import time as _time
@@ -1276,7 +1300,7 @@ def _yield_(x):
     return x
 
 '''.lstrip()
-    BASE_PREFIX_LINES = 14
+    BASE_PREFIX_LINES = 13
 
     prefix = BASE_PREFIX
     prefix_lines = BASE_PREFIX_LINES
@@ -1460,25 +1484,37 @@ class MainMenu(tk.Menu):
         self.add_cascade(label = 'Images', menu = self.images_dropdown)
 
         self.run_menu_entries = {}
+        self.run_menu_seps = 0
         self.run_menu = tk.Menu(self, **MENU_STYLE)
         def add_run_menu_command(name: str, **kwargs):
-            self.run_menu_entries[name] = len(self.run_menu_entries)
+            self.run_menu_entries[name] = len(self.run_menu_entries) + self.run_menu_seps
             self.run_menu.add_command(**kwargs)
+        def add_run_menu_sep():
+            self.run_menu.add_separator()
+            self.run_menu_seps += 1
 
+        def copy_pub_id():
+            root.clipboard_clear()
+            root.clipboard_append(self.public_id)
         add_run_menu_command('run-project', label = 'Run Project', command = play_button, accelerator = 'F5')
         add_run_menu_command('stop-project', label = 'Stop Project', command = play_button, state = tk.DISABLED)
+        add_run_menu_sep()
+        add_run_menu_command('copy-pub-id', label = 'Copy Public ID', command = copy_pub_id)
         self.add_cascade(label = 'Run', menu = self.run_menu)
 
         root.bind_all('<F5>', lambda e: play_button())
 
     @property
+    def public_id(self):
+        return f'{self._project_name}@{content.project.project_id}'
+    @property
     def project_path(self):
         return self._project_path
     @project_path.setter
     def project_path(self, p):
         self._project_path = p
         self._project_name = 'untitled' if p is None else basename_noext(p)
-        root.title(f'NetsBlox-Python - {self._project_name}@{content.project.project_id} ({"unsaved" if p is None else p})')
+        root.title(f'NetsBlox-Python - {self.public_id} ({"unsaved" if p is None else p})')
     @property
     def project_name(self):
         return self._project_name
